from rpytest.ft import ClassWithFields


def test_fields():

    c = ClassWithFields()

    assert "should_be_ignored" not in dir(c)
    assert "array_of_two" in dir(c)

    assert len(c.array_of_two) == 2
    assert c.array_of_two[0] == 0x10
    assert c.array_of_two[1] == 0x22

    c.array_of_two[1] = 0x42
    assert c.get_array_of_two(1) == 0x42

    assert c.const_field == 3

    assert ClassWithFields.static_int == 4
    ClassWithFields.static_int = 44
    assert ClassWithFields.static_int == 44

    assert ClassWithFields.static_const == 5
    assert ClassWithFields.static_constexpr == 6

    assert c.actual_int == 2
    assert c.ref_int == 2

    c.actual_int = 3
    assert c.ref_int == 3

    # TODO: this probably has subtle issues, should always
    #       be readonly?
    c.ref_int = 4
    assert c.actual_int == 4



from keyword import iskeyword
import re
import sphinxify
import typing

from .hooks_datacfg import (
    HooksDataYaml,
    BufferType,
    ClassData,
    EnumValue,
    FunctionData,
    PropData,
    PropAccess,
    ReturnValuePolicy,
)
from .generator_data import GeneratorData, MissingReporter
from .mangle import trampoline_signature


# TODO: this isn't the best solution
def _gen_int_types():
    for i in ("int", "uint"):
        for j in ("", "_fast", "_least"):
            for k in ("8", "16", "32", "64"):
                yield f"{i}{j}{k}_t"
    yield "intmax_t"
    yield "uintmax_t"


_int32_types = frozenset(_gen_int_types())


_rvp_map = {
    ReturnValuePolicy.TAKE_OWNERSHIP: ", py::return_value_policy::take_ownership",
    ReturnValuePolicy.COPY: ", py::return_value_policy::copy",
    ReturnValuePolicy.MOVE: ", py::return_value_policy::move",
    ReturnValuePolicy.REFERENCE: ", py::return_value_policy::reference",
    ReturnValuePolicy.REFERENCE_INTERNAL: ", py::return_value_policy::reference_internal",
    ReturnValuePolicy.AUTOMATIC: "",
    ReturnValuePolicy.AUTOMATIC_REFERENCE: ", py::return_value_policy::automatic_reference",
}

# fmt: off
_operators = {
    # binary
    "-", "+", "*", "/", "%", "&", "^", "==", "!=", "|", ">", ">=", "<", "<=",
    # inplace
    "+=", "-=", "*=", "/=", "%=", "&=", "^=", "|=",
}
# fmt: on

_type_caster_seps = re.compile(r"[<>\(\)]")


class HookError(Exception):
    pass


def _using_signature(fn):
    return f"{fn['parent']['x_qualname_']}_{fn['name']}"


class Hooks:
    """
    Header2Whatever hooks used for generating C++ wrappers
    """

    _qualname_bad = ":<>="
    _qualname_trans = str.maketrans(_qualname_bad, "_" * len(_qualname_bad))

    def __init__(
        self,
        data: HooksDataYaml,
        casters: typing.Dict[str, typing.Dict[str, typing.Any]],
        report_only: bool,
    ):
        self.gendata = GeneratorData(data)
        self.rawdata = data
        self.casters = casters
        self.report_only = report_only
        self.has_operators = False
        self.has_vcheck = False

        self.types: typing.Set[str] = set()
        self.class_hierarchy: typing.Dict[str, typing.List[str]] = {}

        self.subpackages: typing.Dict[str, str] = {}

    def report_missing(self, name: str, reporter: MissingReporter):
        self.gendata.report_missing(name, reporter)

    def _add_type_caster(self, typename: str):
        # defer until the end since there's lots of duplication
        self.types.add(typename)

    def _add_subpackage(self, v, data):
        if data.subpackage:
            var = "pkg_" + data.subpackage.replace(".", "_")
            self.subpackages[data.subpackage] = var
            v["x_module_var"] = var
        else:
            v["x_module_var"] = "m"

    def _get_type_caster_cfgs(self, typename: str):
        tmpl_idx = typename.find("<")
        if tmpl_idx == -1:
            typenames = [typename]
        else:
            typenames = [typename[:tmpl_idx]] + _type_caster_seps.split(
                typename[tmpl_idx:].replace(" ", "")
            )
        for typename in typenames:
            if typename:
                ccfg = self.casters.get(typename)
                if ccfg:
                    yield ccfg

    def _get_type_caster_includes(self):
        includes = set()
        for typename in self.types:
            for ccfg in self._get_type_caster_cfgs(typename):
                includes.add(ccfg["hdr"])
        return sorted(includes)

    def _set_name(self, name, data, strip_prefixes=None, is_operator=False):
        if data.rename:
            return data.rename

        if strip_prefixes is None:
            strip_prefixes = self.rawdata.strip_prefixes

        if strip_prefixes:
            for pfx in strip_prefixes:
                if name.startswith(pfx):
                    n = name[len(pfx) :]
                    if n.isidentifier():
                        name = n
                        break

        if iskeyword(name):
            return f"{name}_"
        if not name.isidentifier() and not is_operator:
            if not self.report_only:
                raise ValueError(f"name {name!r} is not a valid identifier")

        return name

    def _process_doc(
        self, thing, data, append_prefix="", param_remap: typing.Dict[str, str] = {}
    ) -> typing.Optional[typing.List[str]]:
        doc = ""

        if data.doc is not None:
            doc = data.doc
        elif "doxygen" in thing:
            doc = thing["doxygen"]
            if param_remap:
                d = sphinxify.Doc.from_comment(doc)
                for param in d.params:
                    new_name = param_remap.get(param.name)
                    if new_name:
                        param.name = new_name
                doc = str(d)
            else:
                doc = sphinxify.process_raw(doc)

        if data.doc_append is not None:
            doc += f"\n{append_prefix}" + data.doc_append.replace(
                "\n", f"\n{append_prefix}"
            )

        return self._quote_doc(doc)

    def _quote_doc(
        self, doc: typing.Optional[str]
    ) -> typing.Optional[typing.List[str]]:
        doc_quoted: typing.Optional[typing.List[str]] = None
        if doc:
            # TODO
            doc = doc.replace("\\", "\\\\").replace('"', '\\"')
            doc_quoted = doc.splitlines(keepends=True)
            doc_quoted = ['"%s"' % (dq.replace("\n", "\\n"),) for dq in doc_quoted]

        return doc_quoted

    def _resolve_default(self, fn, p, name):
        if isinstance(name, (int, float)):
            return str(name)
        if name in ("NULL", "nullptr"):
            return name

        if name and name[0] == "{" and name[-1] == "}":
            if p["array"]:
                return name
            return f"{p['x_type']}{name}"

        # if there's a parent, look there
        parent = fn["parent"]
        if parent:
            for prop in parent["properties"]["public"]:
                if prop["name"] == name:
                    name = f"{parent['namespace']}::{parent['name']}::{name}"
        return name

    def _maybe_add_default_arg_cast(self, p, name):
        if not p.get("disable_type_caster_default_cast", False):
            found_typename = None
            for ccfg in self._get_type_caster_cfgs(p["x_type"]):
                if ccfg.get("darg"):
                    if found_typename and found_typename != ccfg["typename"]:
                        raise HookError(
                            f"multiple type casters found for {p['name']} ({p['x_type']}), use disable_type_caster_default_cast"
                        )
                    found_typename = ccfg["typename"]
                    name = f"({found_typename}){name}"

        return name

    def _get_function_signature(self, fn):
        param_sig = ", ".join(
            p.get("enum", p["raw_type"]) + "&" * p["reference"] + "*" * p["pointer"]
            for p in fn["parameters"]
        )
        param_sig = param_sig.replace(" >", ">")
        if fn["const"]:
            if param_sig:
                param_sig += " [const]"
            else:
                param_sig = "[const]"

        return param_sig

    def _process_base_param(self, decl_param):
        params = decl_param.get("params")
        if params:
            # recurse
            params = [self._process_base_param(param) for param in params]
            return f"{decl_param['param']}<{', '.join(params)}>"
        else:
            return decl_param["param"]

    def _make_base_params(
        self, base_decl_params, pybase_params: typing.Set[str]
    ) -> str:
        base_params = [
            self._process_base_param(decl_param) for decl_param in base_decl_params
        ]

        for decl_param in base_params:
            pybase_params.add(decl_param)

        return ", ".join(base_params)

    def _extract_typealias(
        self,
        in_ta: typing.List[str],
        out_ta: typing.List[str],
        ta_names: typing.Set[str],
    ):
        for typealias in in_ta:
            if typealias.startswith("template"):
                out_ta.append(typealias)
            else:
                teq = typealias.find("=")
                if teq != -1:
                    ta_name = typealias[:teq].strip()
                    out_ta.append(f"using {typealias}")
                else:
                    ta_name = typealias.split("::")[-1]
                    out_ta.append(f"using {ta_name} = {typealias}")
                ta_names.add(ta_name)

    def _enum_hook(self, en, enum_data):
        ename = en.get("name")
        value_prefix = None
        if ename:
            value_prefix = enum_data.value_prefix
            if not value_prefix:
                value_prefix = ename

            en["x_name"] = self._set_name(ename, enum_data)

        en["x_doc_quoted"] = self._process_doc(en, enum_data)

        if value_prefix:
            strip_prefixes = [value_prefix + "_", value_prefix]
        else:
            strip_prefixes = []

        for v in en["values"]:
            name = v["name"]
            v_data = enum_data.values.get(name)
            if v_data is None:
                v_data = EnumValue()
            v["x_name"] = self._set_name(name, v_data, strip_prefixes)
            v["data"] = v_data
            v["x_doc_quoted"] = self._process_doc(v, v_data, append_prefix="  ")

    def header_hook(self, header, data):
        """Called for each header"""
        data["trampoline_signature"] = trampoline_signature
        data["using_signature"] = _using_signature

        for en in header.enums:
            en["x_namespace"] = en["namespace"]
            enum_data = self.gendata.get_enum_data(en.get("name"))
            en["data"] = enum_data

            self._add_subpackage(en, enum_data)
            self._enum_hook(en, enum_data)

        for v in header.variables:
            var_data = self.gendata.get_prop_data(v["name"])
            v["data"] = var_data
            self._add_type_caster(v["raw_type"])

        for _, u in header.using.items():
            self._add_type_caster(u["raw_type"])

        data["class_hierarchy"] = self.class_hierarchy
        data["subpackages"] = self.subpackages
        data["x_has_operators"] = self.has_operators

        templates = {}
        for k, tmpl_data in data["data"].templates.items():
            qualname = tmpl_data.qualname
            if "::" not in qualname:
                qualname = f"::{qualname}"
            tmpl_data_d = tmpl_data.dict()
            tmpl_data_d["x_qualname_"] = qualname.translate(self._qualname_trans)
            tmpl_data_d["x_doc_set"] = self._quote_doc(tmpl_data.doc)
            doc_add = tmpl_data.doc_append
            if doc_add:
                doc_add = f"\n{doc_add}"
            tmpl_data_d["x_doc_add"] = self._quote_doc(doc_add)
            self._add_subpackage(tmpl_data_d, tmpl_data)
            templates[k] = tmpl_data_d

            for param in tmpl_data.params:
                self._add_type_caster(param)

        data["templates"] = templates

        data["type_caster_includes"] = self._get_type_caster_includes()
        data["has_vcheck"] = self.has_vcheck

        user_typealias = []
        self._extract_typealias(self.rawdata.typealias, user_typealias, set())
        data["user_typealias"] = user_typealias

    def _function_hook(self, fn, data: FunctionData, internal: bool = False):
        """shared with methods/functions"""

        # Python exposed function name converted to camelcase
        x_name = self._set_name(fn["name"], data, is_operator=fn.get("operator", False))
        if not data.rename and not x_name[:2].isupper():
            x_name = x_name[0].lower() + x_name[1:]

        # if cpp_code is specified, don't release the gil unless the user
        # specifically asks for it
        if data.no_release_gil is None:
            if data.cpp_code:
                data.no_release_gil = True

        x_in_params = []
        x_out_params = []
        x_all_params = []
        x_rets = []
        x_temps = []
        x_keepalives = []

        param_remap = {}

        x_genlambda = False
        x_lambda_pre: typing.List[str] = []
        x_lambda_post: typing.List[str] = []

        # Use this if one of the parameter types don't quite match
        param_override = data.param_override

        # buffers: accepts a python object that supports the buffer protocol
        #          as input. If the buffer is an 'out' buffer, then it
        #          will request a writeable buffer. Data is written by the
        #          wrapped function to that buffer directly, and the length
        #          written (if the length is a pointer) will be returned
        buffer_params = {}
        buflen_params = {}
        if data.buffers:
            for bufinfo in data.buffers:
                if bufinfo.src == bufinfo.len:
                    raise ValueError(
                        f"buffer src({bufinfo.src}) and len({bufinfo.len}) cannot be the same"
                    )
                buffer_params[bufinfo.src] = bufinfo
                buflen_params[bufinfo.len] = bufinfo

        self._add_type_caster(fn["returns"])

        is_constructor = fn.get("constructor")

        for i, p in enumerate(fn["parameters"]):

            if is_constructor and p["reference"] == 1:
                x_keepalives.append((1, i + 2))

            if p["raw_type"] in _int32_types:
                p["fundamental"] = True
                p["unresolved"] = False

            if p["name"] == "":
                p["name"] = "param%s" % i
            p["x_type"] = p.get("enum", p["raw_type"])

            if p["pointer"]:
                p["x_callname"] = p["name"]
            elif p["reference"]:
                p["x_callname"] = f"std::forward<decltype({p['name']})>({p['name']})"
            else:
                p["x_callname"] = f"std::move({p['name']})"
            # this is used in trampoline functions
            # - separate from x_callname because it's just a passthrough
            p["x_virtual_callname"] = p["x_callname"]

            orig_pname = p["name"]
            p["x_retname"] = orig_pname

            po = param_override.get(orig_pname)
            if po:
                p.update(po.dict(exclude_unset=True))

            py_pname = p["name"]
            if iskeyword(py_pname):
                py_pname = f"{py_pname}_"

            if orig_pname != py_pname:
                param_remap[orig_pname] = py_pname

            p["x_pyarg"] = f'py::arg("{py_pname}")'

            if "default" in p:
                default = self._resolve_default(fn, p, p["default"])
                default = self._maybe_add_default_arg_cast(p, default)
                p["default"] = default
                if default:
                    p["x_pyarg"] += "=" + p["default"]

            ptype = "in"

            buflen = buflen_params.pop(p["name"], None)

            if p["name"] in buffer_params:
                bufinfo = buffer_params.pop(p["name"])
                x_genlambda = True
                bname = f"__{bufinfo.src}"
                p["constant"] = 1
                p["reference"] = 1
                p["pointer"] = 0

                p["x_callname"] = f"({p['x_type']}*){bname}.ptr"
                p["x_type"] = "py::buffer"

                # this doesn't seem to be true for bytearrays, which is silly
                # x_lambda_pre.append(
                #     f'if (PyBuffer_IsContiguous((Py_buffer*){p["name"]}.ptr(), \'C\') == 0) throw py::value_error("{p["name"]}: buffer must be contiguous")'
                # )

                # TODO: check for dimensions, strides, other dangerous things

                # bufinfo was validated and converted before it got here
                if bufinfo.type is BufferType.IN:
                    ptype = "in"
                    x_lambda_pre += [f"auto {bname} = {p['name']}.request(false)"]
                else:
                    ptype = "in"
                    x_lambda_pre += [f"auto {bname} = {p['name']}.request(true)"]

                x_lambda_pre += [f"{bufinfo.len} = {bname}.size * {bname}.itemsize"]

                if bufinfo.minsz:
                    x_lambda_pre.append(
                        f'if ({bufinfo.len} < {bufinfo.minsz}) throw py::value_error("{p["name"]}: minimum buffer size is {bufinfo.minsz}")'
                    )

            elif buflen:
                if p["pointer"]:
                    p["x_callname"] = f"&{buflen.len}"
                    ptype = "out"
                else:
                    # if it's not a pointer, then the called function
                    # can't communicate through it, so ignore the parameter
                    p["x_callname"] = buflen.len
                    x_temps.append(p)
                    ptype = "ignored"

            elif p.get("force_out") or (
                (p["pointer"] or p["reference"] == 1)
                and not p["constant"]
                and p["fundamental"]
            ):
                if p["pointer"]:
                    p["x_callname"] = f"&{p['x_callname']}"
                else:
                    p["x_callname"] = p["name"]
                ptype = "out"
            elif p["array"]:
                asz = p.get("array_size", 0)
                if asz:
                    p["x_type"] = "std::array<%s, %s>" % (p["x_type"], asz)
                    p["x_callname"] = "%(x_callname)s.data()" % p
                    if "default" not in p:
                        p["default"] = "{}"
                else:
                    # it's a vector
                    pass
                ptype = "out"

            if p.get("ignore"):
                pass
            else:
                x_all_params.append(p)
                if ptype == "out":
                    x_out_params.append(p)
                    x_temps.append(p)
                elif ptype == "in":
                    x_in_params.append(p)

            self._add_type_caster(p["x_type"])

            if p["constant"]:
                p["x_type"] = "const " + p["x_type"]

            p["x_type_full"] = p["x_type"]
            p["x_type_full"] += "&" * p["reference"]
            p["x_type_full"] += "*" * p["pointer"]

            p["x_decl"] = "%s %s" % (p["x_type_full"], p["name"])

        if buffer_params:
            raise ValueError(
                "incorrect buffer param names '%s'"
                % ("', '".join(buffer_params.keys()))
            )

        x_callstart = ""
        x_callend = ""
        x_wrap_return = ""
        x_return_value_policy = _rvp_map[data.return_value_policy]

        if x_out_params:
            x_genlambda = True

            # Return all out parameters
            x_rets.extend(x_out_params)

        if fn["rtnType"] != "void":
            x_callstart = "auto __ret ="
            x_rets.insert(0, dict(x_retname="__ret", x_type=fn["rtnType"]))

        if len(x_rets) == 1 and x_rets[0]["x_type"] != "void":
            x_wrap_return = "return %s;" % x_rets[0]["x_retname"]
        elif len(x_rets) > 1:
            x_wrap_return = "return std::make_tuple(%s);" % ",".join(
                [p["x_retname"] for p in x_rets]
            )

        # Temporary values to store out parameters in
        if x_temps:
            for out in reversed(x_temps):
                odef = out.get("default", "0")
                if not odef:
                    x_lambda_pre.insert(0, f"{out['x_type']} {out['name']}")
                elif odef.startswith("{"):
                    x_lambda_pre.insert(0, f"{out['x_type']} {out['name']}{odef}")
                else:
                    x_lambda_pre.insert(0, f"{out['x_type']} {out['name']} = {odef}")

        # Rename functions
        if data.rename:
            x_name = data.rename
        elif data.internal or internal:
            x_name = "_" + x_name
        elif fn["constructor"]:
            x_name = "__init__"

        doc_quoted = self._process_doc(fn, data, param_remap=param_remap)

        if data.keepalive is not None:
            x_keepalives = data.keepalive

        if not self.report_only:
            if fn["template"]:
                if data.template_impls is None and not data.cpp_code:
                    raise ValueError(
                        f"{fn['name']}: must specify template impls for function template"
                    )
            else:
                if data.template_impls is not None:
                    raise ValueError(
                        f"{fn['name']}: cannot specify template_impls for non-template functions"
                    )

            if data.ignore_pure and not fn["pure_virtual"]:
                raise ValueError(
                    f"{fn['name']}: cannot specify ignore_pure for function that isn't pure"
                )

            if data.trampoline_cpp_code and (not fn["override"] and not fn["virtual"]):
                raise ValueError(
                    f"{fn['name']}: cannot specify trampoline_cpp_code for a non-virtual method"
                )

            if data.virtual_xform and (not fn["override"] and not fn["virtual"]):
                raise ValueError(
                    f"{fn['name']}: cannot specify virtual_xform for a non-virtual method"
                )

            if fn.get("ref_qualifiers", "") == "&&":
                # pybind11 doesn't support this, user must fix it
                if not data.ignore_py and not data.cpp_code:
                    raise ValueError(
                        f"{fn['name']}: has && ref-qualifier which cannot be directly bound by pybind11, must specify cpp_code or ignore_py"
                    )

        # if "hook" in data:
        #     eval(data["hook"])(fn, data)

        # bind new attributes to the function definition
        # -> previously used locals(), but this is more explicit
        #    and easier to not mess up
        fn.update(
            dict(
                data=data,
                # transforms
                x_name=x_name,
                x_all_params=x_all_params,
                x_in_params=x_in_params,
                x_out_params=x_out_params,
                x_rets=x_rets,
                x_keepalives=x_keepalives,
                x_return_value_policy=x_return_value_policy,
                # lambda generation
                x_genlambda=x_genlambda,
                x_callstart=x_callstart,
                x_lambda_pre=x_lambda_pre,
                x_lambda_post=x_lambda_post,
                x_callend=x_callend,
                x_wrap_return=x_wrap_return,
                # docstrings
                x_doc_quoted=doc_quoted,
            )
        )

    def function_hook(self, fn, data):
        if fn.get("operator"):
            fn["data"] = FunctionData(ignore=True)
            return

        signature = self._get_function_signature(fn)
        data = self.gendata.get_function_data(fn, signature)
        if data.ignore:
            fn["data"] = data
            return

        self._add_subpackage(fn, data)
        self._function_hook(fn, data)

    def class_hook(self, cls, data):

        if cls["parent"] is not None and cls["access_in_parent"] == "private":
            cls["data"] = ClassData(ignore=True)
            return

        cls_name = cls["name"]
        cls_key = cls_name
        c = cls
        while c["parent"]:
            c = c["parent"]
            cls_key = c["name"] + "::" + cls_key

        class_data = self.gendata.get_class_data(cls_key)
        cls["data"] = class_data

        if class_data.ignore:
            return

        for _, u in cls["using"].items():
            self._add_type_caster(u["raw_type"])

        for typename in class_data.force_type_casters:
            self._add_type_caster(typename)

        self._add_subpackage(cls, class_data)

        # fix enum paths
        for e in cls["enums"]["public"]:
            e["x_namespace"] = e["namespace"] + "::" + cls_name + "::"
            enum_data = self.gendata.get_cls_enum_data(
                e.get("name"), cls_key, class_data
            )
            e["data"] = enum_data

            self._enum_hook(e, enum_data)

        # update inheritance

        pybase_params = set()

        for base in cls["inherits"]:
            bqual = class_data.base_qualnames.get(base["decl_name"])
            if bqual:
                base["x_class"] = bqual
                # TODO: sometimes need to add this to pybase_params, but
                # that would require parsing this more. Seems sufficiently
                # obscure, going to omit it for now.
                tp = bqual.find("<")
                if tp == -1:
                    base["x_qualname"] = bqual
                    base["x_params"] = ""
                else:
                    base["x_qualname"] = bqual[:tp]
                    base["x_params"] = bqual[tp + 1 : -1]
            else:
                if "::" not in base["decl_name"]:
                    base["x_qualname"] = f'{cls["namespace"]}::{base["decl_name"]}'
                else:
                    base["x_qualname"] = base["decl_name"]

                base_decl_params = base.get("decl_params")
                if base_decl_params:
                    base["x_params"] = self._make_base_params(
                        base_decl_params, pybase_params
                    )
                    base["x_class"] = f"{base['x_qualname']}<{base['x_params']}>"
                else:
                    base["x_params"] = ""
                    base["x_class"] = base["x_qualname"]

            base["x_qualname_"] = base["x_qualname"].translate(self._qualname_trans)

        ignored_bases = {ib: True for ib in class_data.ignored_bases}

        cls["x_inherits"] = [
            base
            for base in cls["inherits"]
            if not ignored_bases.pop(base["class"], None)
            and base["access"] != "private"
        ]
        if not self.report_only and ignored_bases:
            bases = ", ".join(base["class"] for base in cls["inherits"])
            invalid_bases = ", ".join(ignored_bases.keys())
            raise ValueError(
                f"{cls_name}: ignored_bases contains non-existant bases "
                + f"{invalid_bases}; valid bases are {bases}"
            )

        # No template stuff
        simple_cls_qualname = f'{cls["namespace"]}::{cls_name}'

        # Template stuff
        if cls["parent"]:
            cls_qualname = f'{cls["parent"]["x_qualname"]}::{cls_name}'
        else:
            cls_qualname = simple_cls_qualname

        cls["x_qualname_"] = cls_qualname.translate(self._qualname_trans)
        self.class_hierarchy[simple_cls_qualname] = [
            base["x_qualname"] for base in cls["x_inherits"]
        ] + class_data.force_depends

        # <N, .. >
        template_argument_list = ""
        # <typename N, .. >
        template_parameter_list = ""

        if class_data.template_params:
            if class_data.subpackage:
                raise ValueError(
                    f"{cls_name}: classes with subpackages must define subpackage on template instantiation"
                )

            template_args = []
            template_params = []

            base_template_args = []
            base_template_params = []

            for param in class_data.template_params:
                if " " in param:
                    arg = param.split(" ", 1)[1]
                else:
                    arg = param
                    param = f"typename {param}"

                template_args.append(arg)
                template_params.append(param)

                if arg in pybase_params:
                    base_template_args.append(arg)
                    base_template_params.append(param)

            template_argument_list = ", ".join(template_args)
            template_parameter_list = ", ".join(template_params)

            cls_qualname = f"{cls_qualname}<{template_argument_list}>"
        else:
            base_template_params = None
            base_template_args = None

        if base_template_params:
            cls["x_pybase_args"] = ", ".join(base_template_args)
            cls["x_pybase_params"] = ", ".join(base_template_params)
        else:
            cls["x_pybase_args"] = ""
            cls["x_pybase_params"] = ""

        if not self.report_only:
            if "template" in cls:
                if template_parameter_list == "":
                    raise ValueError(
                        f"{cls_name}: must specify template_params for templated class, or ignore it"
                    )
            else:
                if template_parameter_list != "":
                    raise ValueError(
                        f"{cls_name}: cannot specify template_params for non-template class"
                    )

        cls["x_qualname"] = cls_qualname

        has_constructor = False
        is_polymorphic = class_data.is_polymorphic
        vcheck_fns = []
        cls["x_vcheck_fns"] = vcheck_fns

        # bad assumption? yep
        if cls["inherits"]:
            is_polymorphic = True

        for access in ("public", "protected", "private"):

            for fn in cls["methods"][access]:
                if fn["constructor"]:
                    has_constructor = True
                if fn["override"] or fn["virtual"]:
                    is_polymorphic = True

                operator = fn.get("operator")

                # Ignore some operators, move constructors, copy constructors
                if (
                    (operator and operator not in _operators)
                    or fn.get("destructor")
                    or (
                        fn.get("constructor")
                        and fn["parameters"]
                        and fn["parameters"][0]["class"] is cls
                    )
                    or fn["deleted"]
                ):
                    fn["data"] = FunctionData(ignore=True)
                    continue

                is_private = access == "private"

                # this has to be done even on private functions, because
                # we do overload detection here
                signature = self._get_function_signature(fn)
                method_data = self.gendata.get_function_data(
                    fn, signature, cls_key, class_data, is_private
                )

                if not is_private:
                    if method_data.ignore:
                        fn["data"] = method_data
                        continue

                    # If the method has cpp_code defined, it must either match the function
                    # signature of the method, or virtual_xform must be defined with an
                    # appropriate conversion. If neither of these are true, it will lead
                    # to difficult to diagnose errors at runtime. We add a static assert
                    # to try and catch these errors at compile time
                    need_vcheck = (
                        (fn["override"] or fn["virtual"])
                        and method_data.cpp_code
                        and not method_data.virtual_xform
                        and not method_data.trampoline_cpp_code
                        and not cls["final"]
                        and not class_data.force_no_trampoline
                    )
                    if need_vcheck:
                        vcheck_fns.append(fn)
                        self.has_vcheck = True

                    if operator:
                        self.has_operators = True
                        if method_data.no_release_gil is None:
                            method_data.no_release_gil = True

                    internal = access != "public"

                    try:
                        self._function_hook(fn, method_data, internal=internal)
                    except Exception as e:
                        raise HookError(f"{cls_key}::{fn['name']}") from e

        has_trampoline = (
            is_polymorphic and not cls["final"] and not class_data.force_no_trampoline
        )
        for access in ("public", "protected", "private"):
            # class attributes
            for v in cls["properties"][access]:
                if access in "private" or (
                    access == "protected" and not has_trampoline
                ):
                    v["data"] = PropData(ignore=True)
                    continue

                prop_name = v["name"]
                propdata = self.gendata.get_cls_prop_data(
                    prop_name, cls_key, class_data
                )
                self._add_type_caster(v["raw_type"])
                v["data"] = propdata
                if propdata.rename:
                    v["x_name"] = propdata.rename
                else:
                    v["x_name"] = v["name"] if access == "public" else "_" + v["name"]

                if propdata.access == PropAccess.AUTOMATIC:
                    # We assume that a struct intentionally has readwrite data
                    # attributes regardless of type
                    if cls["declaration_method"] != "class":
                        x_readonly = False
                    else:
                        # Properties that aren't fundamental or a reference are readonly unless
                        # overridden by the hook configuration
                        x_readonly = not v["fundamental"] and not v["reference"]
                elif propdata.access == PropAccess.READONLY:
                    x_readonly = True
                else:
                    x_readonly = False

                v["x_readonly"] = x_readonly
                v["x_doc_quoted"] = self._process_doc(v, propdata)

        cls["x_has_trampoline"] = has_trampoline
        cls["x_template_parameter_list"] = template_parameter_list
        if cls["x_has_trampoline"]:
            tmpl = ""
            if template_argument_list:
                tmpl = f", {template_argument_list}"

            trampoline_cfg = f"rpygen::PyTrampolineCfg_{cls['x_qualname_']}<{template_argument_list}>"
            cls[
                "x_trampoline_name"
            ] = f"rpygen::PyTrampoline_{cls['x_qualname_']}<typename {cls_qualname}{tmpl}, typename {trampoline_cfg}>"
            cls["x_trampoline_var"] = f"{cls_name}_Trampoline"
        elif class_data.trampoline_inline_code is not None:
            raise HookError(
                f"{cls_key} has trampoline_inline_code specified, but there is no trampoline!"
            )

        cls["x_has_constructor"] = has_constructor
        cls["x_varname"] = "cls_" + cls_name
        cls["x_name"] = self._set_name(cls_name, class_data)
        cls["x_doc_quoted"] = self._process_doc(cls, class_data)
        cls["x_inline_code"] = class_data.inline_code or ""

        # do logic for extracting user defined typealiases here
        # - these are at class scope, so they can include template
        typealias_names = set()
        x_user_typealias = []
        self._extract_typealias(class_data.typealias, x_user_typealias, typealias_names)

        # autodetect embedded using directives, but don't override anything
        # the user specifies
        # - these are in block scope, so they cannot include templates
        x_auto_typealias = []
        for name, using in cls["using"].items():
            if (
                using["access"] == "public"
                and name not in typealias_names
                and not using["template"]
                and using["using_type"] == "typealias"
            ):
                x_auto_typealias.append(
                    f"using {name} [[maybe_unused]] = typename {cls['x_qualname']}::{name}"
                )

        cls["x_user_typealias"] = x_user_typealias
        cls["x_auto_typealias"] = x_auto_typealias



from rpytest import ft
import pytest


# C++ check functions
getBaseOnly = ft.IBase.getBaseOnly
getBaseAndChild = ft.IBase.getBaseAndChild
getBaseAndPyChild = ft.IBase.getBaseAndPyChild
getBaseAndChildFinal = ft.IBase.getBaseAndChildFinal
getBaseAndGrandchild = ft.IBase.getBaseAndGrandchild


class PyIBase(ft.IBase):
    def baseOnly(self):
        return "pyibase::baseOnly"


def test_inheritance_base():

    # create a base, check its virtual methods
    b = ft.IBase()
    assert b.baseOnly() == "base::baseOnly"
    assert b.baseAndChild() == "base::baseAndChild"
    assert b.baseAndPyChild() == "base::baseAndPyChild"
    assert b.baseAndChildFinal() == "base::baseAndChildFinal"
    assert b.baseAndGrandchild() == "base::baseAndGrandchild"

    # C++ function should see the same thing
    assert getBaseOnly(b) == "base::baseOnly"
    assert getBaseAndChild(b) == "base::baseAndChild"
    assert getBaseAndPyChild(b) == "base::baseAndPyChild"
    assert getBaseAndChildFinal(b) == "base::baseAndChildFinal"
    assert getBaseAndGrandchild(b) == "base::baseAndGrandchild"


def test_inheritance_pybase():

    # Overridden version should be the same
    pyb = PyIBase()
    assert pyb.baseOnly() == "pyibase::baseOnly"
    assert pyb.baseAndChild() == "base::baseAndChild"
    assert pyb.baseAndPyChild() == "base::baseAndPyChild"
    assert pyb.baseAndChildFinal() == "base::baseAndChildFinal"
    assert pyb.baseAndGrandchild() == "base::baseAndGrandchild"

    # C++ function should see the same thing
    assert getBaseOnly(pyb) == "pyibase::baseOnly"
    assert getBaseAndChild(pyb) == "base::baseAndChild"
    assert getBaseAndPyChild(pyb) == "base::baseAndPyChild"
    assert getBaseAndChildFinal(pyb) == "base::baseAndChildFinal"
    assert getBaseAndGrandchild(pyb) == "base::baseAndGrandchild"


def test_inheritance_child():

    # child
    c = ft.IChild()
    assert c.baseOnly() == "base::baseOnly"
    assert c.baseAndChild() == "child::baseAndChild"
    assert c.baseAndPyChild() == "base::baseAndPyChild"
    assert c.baseAndChildFinal() == "child::baseAndChildFinal"
    assert c.baseAndGrandchild() == "base::baseAndGrandchild"

    assert c.getI() == 42

    # C++ function should see the same thing
    assert getBaseOnly(c) == "base::baseOnly"
    assert getBaseAndChild(c) == "child::baseAndChild"
    assert getBaseAndPyChild(c) == "base::baseAndPyChild"
    assert getBaseAndChildFinal(c) == "child::baseAndChildFinal"
    assert getBaseAndGrandchild(c) == "base::baseAndGrandchild"


class PyIChild(ft.IChild):
    def baseAndChild(self):
        return "pyichild::baseAndChild"

    def baseAndPyChild(self):
        return "pyichild::baseAndPyChild"

    # baseAndChildFinal is final, so even though it's overridden here, it won't
    # work in C++ land
    # .. would be nice if the pybind11 metaclass detected that, but probably
    #    will never happen
    def baseAndChildFinal(self):
        return "pyichild::baseAndChildFinal"


def test_inheritance_pychild():
    # child
    pyc = PyIChild()
    assert pyc.baseOnly() == "base::baseOnly"
    assert pyc.baseAndChild() == "pyichild::baseAndChild"
    assert pyc.baseAndPyChild() == "pyichild::baseAndPyChild"
    assert pyc.baseAndChildFinal() == "pyichild::baseAndChildFinal"
    assert pyc.baseAndGrandchild() == "base::baseAndGrandchild"

    assert pyc.getI() == 42

    # C++ function should see the same thing, except the final function is broken
    assert getBaseOnly(pyc) == "base::baseOnly"
    assert getBaseAndChild(pyc) == "pyichild::baseAndChild"
    assert getBaseAndPyChild(pyc) == "pyichild::baseAndPyChild"
    assert getBaseAndChildFinal(pyc) == "child::baseAndChildFinal"
    assert getBaseAndGrandchild(pyc) == "base::baseAndGrandchild"


def test_inheritance_grandchild():
    # grandchild
    g = ft.IGrandChild()
    assert g.baseOnly() == "base::baseOnly"
    assert g.baseAndChild() == "child::baseAndChild"
    assert g.baseAndPyChild() == "base::baseAndPyChild"
    assert g.baseAndChildFinal() == "child::baseAndChildFinal"
    assert g.baseAndGrandchild() == "grandchild::baseAndGrandchild"

    # C++ function should see the same thing
    assert getBaseOnly(g) == "base::baseOnly"
    assert getBaseAndChild(g) == "child::baseAndChild"
    assert getBaseAndPyChild(g) == "base::baseAndPyChild"
    assert getBaseAndChildFinal(g) == "child::baseAndChildFinal"
    assert getBaseAndGrandchild(g) == "grandchild::baseAndGrandchild"


def test_inheritance_pygrandchild():

    # grandchild is final, so we cannot inherit from it
    with pytest.raises(TypeError):

        class PyGrandChild(ft.IGrandChild):
            pass


def test_inheritance_mchild():

    # child
    c = ft.IMChild()
    assert c.baseOnly() == "base::baseOnly"
    assert c.baseAndChild() == "mchild::baseAndChild"
    assert c.baseAndChildFinal() == "mchild::baseAndChildFinal"
    assert c.baseAndGrandchild() == "base::baseAndGrandchild"

    # assert c.getI() == 42

    # C++ function should see the same thing
    assert getBaseOnly(c) == "base::baseOnly"
    assert getBaseAndChild(c) == "mchild::baseAndChild"
    assert getBaseAndChildFinal(c) == "mchild::baseAndChildFinal"
    assert getBaseAndGrandchild(c) == "base::baseAndGrandchild"


class PyIMChild(ft.IMChild):
    def baseAndChild(self):
        return "pymchild::baseAndChild"

    def baseAndPyChild(self):
        return "pymchild::baseAndPyChild"

    # baseAndChildFinal is final, so even though it's overridden here, it won't
    # work in C++ land
    # .. would be nice if the pybind11 metaclass detected that, but probably
    #    will never happen
    def baseAndChildFinal(self):
        return "pymchild::baseAndChildFinal"


def test_inheritance_pymchild():
    # child
    pyc = PyIMChild()
    assert pyc.baseOnly() == "base::baseOnly"
    assert pyc.baseAndChild() == "pymchild::baseAndChild"
    assert pyc.baseAndPyChild() == "pymchild::baseAndPyChild"
    assert pyc.baseAndChildFinal() == "pymchild::baseAndChildFinal"
    assert pyc.baseAndGrandchild() == "base::baseAndGrandchild"

    # C++ function should see the same thing, except the final function can't be overridden by python
    assert getBaseOnly(pyc) == "base::baseOnly"
    assert getBaseAndChild(pyc) == "pymchild::baseAndChild"
    assert getBaseAndPyChild(pyc) == "pymchild::baseAndPyChild"
    assert getBaseAndChildFinal(pyc) == "mchild::baseAndChildFinal"
    assert getBaseAndGrandchild(pyc) == "base::baseAndGrandchild"


def test_inheritance_usingparent():
    base = ft.UPBase()
    assert base.get5() == 5

    child = ft.UPChild()
    # not automatically added, but at least it's not a compile error either
    # assert child.get5() == 5
    assert child.get6() == 6



from rpytest import ft


def test_basic_template():
    """template/basic.h"""
    s = ft.TBasicString()

    s.setT("string")
    assert s.t == "string"
    assert s.getT() == "string"


def test_dependent_using():
    du = ft.TDependentUsingInt()
    assert du.getThird([1, 2, 3]) == 3


def test_dependent_using2():
    du = ft.TDependentUsing2Int()
    assert du.getThird([1, 2, 3]) == 3


def test_classwithfn():
    assert ft.TClassWithFn.getT(1) == 1
    assert ft.TClassWithFn.getT(False) is False


def test_nested_template():
    """template/nested.h"""
    i = ft.TOuter.Inner()
    assert type(i.t) == int


def test_numeric():
    """template/numeric.h"""
    b4 = ft.TBaseGetN4()
    assert b4.getIt() == 4

    b6 = ft.TBaseGetN6()
    assert b6.getIt() == 6

    c4 = ft.TChildGetN4()
    assert c4.getIt() == 4

    c6 = ft.TChildGetN6()
    assert c6.getIt() == 6


#
# CRTP handling tests
#


def test_crtp_base():
    b = ft.TBase()
    assert b.baseFn() == 42
    assert b.get() == "TBase"


def test_crtp_concrete():
    c = ft.TConcrete()
    assert c.concrete() == 32
    assert c.baseFn() == 42
    assert c.get() == "TCrtp"



import argparse
import glob
import inspect
from os.path import basename, dirname, exists, join, relpath
from pathlib import Path, PurePosixPath
import posixpath
import pprint
import subprocess
import sys
import re
from urllib.request import Request, urlopen
from urllib.error import HTTPError
from collections import defaultdict
import tomli
import tomli_w
from contextlib import suppress

from .setup import Setup
from .generator_data import MissingReporter
from .command.util import get_build_temp_path

from . import overrides
from . import platforms


def get_setup() -> Setup:
    s = Setup()
    s.prepare()

    temp_path = join(get_build_temp_path(), "dlstatic")
    for static_lib in s.static_libs:
        static_lib.set_root(temp_path)

    return s


class BuildDep:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "build-dep", help="Install build dependencies", parents=[parent_parser]
        )
        parser.add_argument("--install", help="Actually do it", action="store_true")
        parser.add_argument(
            "--pre",
            action="store_true",
            default=False,
            help="Include pre-release and development versions.",
        )
        parser.add_argument("--find-links", help="Find links arg", default=None)
        return parser

    def run(self, args):
        s = get_setup()
        requirements = s.pyproject.get("build-system", {}).get("requires", [])
        requirements.extend(s.setup_kwargs.get("install_requires", ""))
        requirements.append("wheel")

        pipargs = [
            sys.executable,
            "-m",
            "pip",
            "--disable-pip-version-check",
            "install",
        ]
        if args.pre:
            pipargs.append("--pre")
        if args.find_links:
            pipargs.extend(
                [
                    "--find-links",
                    args.find_links,
                ]
            )

        pipargs.extend(requirements)
        print(" ".join(pipargs))

        if args.install:
            p = subprocess.run(pipargs)
            return p.returncode


class GenCreator:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "create-gen",
            help="Create YAML files from parsed header files",
            parents=[parent_parser],
        )
        parser.add_argument(
            "--write", help="Write to files if they don't exist", action="store_true"
        )
        parser.add_argument("--strip-prefixes", action="append")

        return parser

    def run(self, args):

        pfx = ""
        if args.strip_prefixes:
            pfx = "strip_prefixes:\n- " + "\n- ".join(args.strip_prefixes) + "\n\n"

        s = get_setup()
        for wrapper in s.wrappers:
            reporter = MissingReporter()
            wrapper.on_build_gen("", reporter)

            nada = True
            for name, report in reporter.as_yaml():
                report = f"---\n\n{pfx}{report}"

                nada = False
                if args.write:
                    if not exists(name):
                        print("Writing", name)
                        with open(name, "w") as fp:
                            fp.write(report)
                    else:
                        print(name, "already exists!")

                print("===", name, "===")
                print(report)

            if nada:
                print("Nothing to do!")


class HeaderScanner:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "scan-headers",
            help="Generate a list of headers in TOML form",
            parents=[parent_parser],
        )
        parser.add_argument("--only-missing", default=False, action="store_true")
        return parser

    def run(self, args):
        s = get_setup()

        already_present = {}
        if args.only_missing:
            for i, wrapper in enumerate(s.project.wrappers.values()):
                files = set()
                if wrapper.autogen_headers:
                    files |= {
                        PurePosixPath(f) for f in wrapper.autogen_headers.values()
                    }
                if wrapper.type_casters:
                    files |= {PurePosixPath(tc.header) for tc in wrapper.type_casters}
                if not files:
                    continue
                for incdir in s.wrappers[i]._generation_search_path():
                    ifiles = already_present.setdefault(incdir, set())
                    incdir = Path(incdir)
                    for f in files:
                        if (incdir / f).exists():
                            ifiles.add(f)

        for wrapper in s.wrappers:
            print(
                f'[tool.robotpy-build.wrappers."{wrapper.package_name}".autogen_headers]'
            )

            # This uses the direct include directories instead of the generation
            # search path as we only want to output a file once
            for incdir in wrapper.get_include_dirs():

                wpresent = already_present.get(incdir, set())

                files = list(
                    sorted(
                        PurePosixPath(relpath(f, incdir))
                        for f in glob.glob(join(incdir, "**", "*.h"), recursive=True)
                        if "rpygen" not in f
                    )
                )

                lastdir = None
                for f in files:
                    if f in wpresent:
                        continue

                    thisdir = f.parent
                    if lastdir is None:
                        if thisdir:
                            print("#", thisdir)
                    elif lastdir != thisdir:
                        print()
                        if thisdir:
                            print("#", thisdir)
                    lastdir = thisdir

                    base = f.stem
                    print(f'{base} = "{f}"')
                print()


class ImportCreator:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "create-imports",
            help="Generate suitable imports for a module",
            parents=[parent_parser],
        )
        parser.add_argument("base", help="Ex: wpiutil")
        parser.add_argument("compiled", nargs="?", help="Ex: wpiutil._impl.wpiutil")
        return parser

    def _rel(self, base: str, compiled: str) -> str:
        base = posixpath.join(*base.split("."))
        compiled = posixpath.join(*compiled.split("."))
        elems = posixpath.relpath(compiled, base).split("/")
        elems = ["" if e == ".." else e for e in elems]
        return f".{'.'.join(elems)}"

    def run(self, args):
        # Runtime Dependency Check
        try:
            import black
        except:
            print("Error, The following module is required to run this tool: black")
            exit(1)

        compiled = args.compiled
        if not compiled:
            compiled = f"{args.base}._{args.base.split('.')[-1]}"

        # TODO: could probably generate this from parsed code, but seems hard
        ctx = {}
        exec(f"from {compiled} import *", {}, ctx)

        relimport = self._rel(args.base, compiled)

        stmt_compiled = "" if not args.compiled else f" {args.compiled}"

        stmt = inspect.cleandoc(
            f"""

            # autogenerated by 'robotpy-build create-imports {args.base}{stmt_compiled}'
            from {relimport} import {','.join(sorted(ctx.keys()))}
            __all__ = ["{'", "'.join(sorted(ctx.keys()))}"]
        
        """
        )

        print(
            subprocess.check_output(
                ["black", "-", "-q"], input=stmt.encode("utf-8")
            ).decode("utf-8")
        )


class PlatformInfo:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "platform-info",
            help="Displays platform-specific information",
            parents=[parent_parser],
        )
        parser.add_argument("--list", action="store_true", default=False)
        parser.add_argument("platform", nargs="?", default=None)
        return parser

    def run(self, args):

        if args.list:
            for name in platforms.get_platform_names():
                print(name)
        else:

            p = platforms.get_platform(args.platform)
            print("platform:")
            pprint.pprint(p)
            print()

            print("override keys:")
            pprint.pprint(platforms.get_platform_override_keys(p))


class ShowOverrides:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "show-override",
            help="Displays pyproject.toml after override processing",
            parents=[parent_parser],
        )
        parser.add_argument("toml", nargs="?", default="pyproject.toml")
        parser.add_argument(
            "-p",
            "--platform",
            default=None,
            help="Use robotpy-build platform-info --list for available platforms",
        )
        return parser

    def run(self, args):

        p = platforms.get_platform(args.platform)
        override_keys = platforms.get_platform_override_keys(p)

        with open(args.toml, "rb") as fp:
            d = tomli.load(fp)

        overrides.apply_overrides(d, override_keys)
        print(tomli_w.dumps(d))


class MavenParser:

    after_archs = [
        "static",
        "debug",
        "staticdebug",
    ]

    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "parse-maven",
            help="Find supported platforms from a pyproject.toml",
            parents=[parent_parser],
        )
        parser.add_argument(
            "toml_link",
            help="Ex: pyproject.toml",
        )
        parser.add_argument(
            "-b",
            "--brute_force",
            action="store_true",
            help="Try known os and arch combinations instead of parsing html (needed for rev)",
        )
        return parser

    def check_url_exists(self, file_url):
        with suppress(Exception):
            if urlopen(Request(file_url)).code == 200:
                return True
        return False

    def run(self, args):

        self.os_names = set()
        self.arch_names = set()
        for plat in platforms._platforms.values():
            self.os_names.add(plat.os)
            self.arch_names.add(plat.arch)

        with open(args.toml_link, "rb") as fp:
            config = tomli.load(fp)["tool"]["robotpy-build"]

            wrappers = {}
            if "wrappers" in config:
                wrappers = {
                    k: v
                    for (k, v) in config["wrappers"].items()
                    if "maven_lib_download" in v
                }

            static_libs = {}
            if "static_libs" in config:
                static_libs = {
                    k: v
                    for (k, v) in config["static_libs"].items()
                    if "maven_lib_download" in v
                }

            if wrappers == {} and static_libs == {}:
                print("No maven_lib_downloads in pyproject.toml")
                exit()

            for w_name, wrapper in {**wrappers, **static_libs}.items():

                if "maven_lib_download" not in wrapper:
                    continue

                mvl = wrapper["maven_lib_download"]

                repo_url = mvl["repo_url"]
                grp = mvl["group_id"].replace(".", "/")
                art = mvl["artifact_id"]
                ver = mvl["version"]

                dir_url = f"{repo_url}/{grp}/{art}/{ver}/"

                plats = defaultdict(set)

                found_source = False
                source_name = None

                if args.brute_force:

                    for os in self.os_names:
                        for arch in self.arch_names:
                            for after_arch in self.after_archs + [""]:
                                classifier = os + arch + after_arch
                                file_url = f"{dir_url}{art}-{ver}-{classifier}.zip"

                                if self.check_url_exists(file_url):
                                    plats[os].add(arch)

                    if self.check_url_exists(f"{dir_url}{art}-{ver}-source.zip"):
                        found_source = True
                        source_name = "source"

                    if self.check_url_exists(f"{dir_url}{art}-{ver}-sources.zip"):
                        found_source = True
                        source_name = "sources"

                else:
                    try:
                        html = str(urlopen(Request(dir_url)).read())
                    except HTTPError as e:
                        if e.code != 404:
                            raise e
                        else:
                            print(
                                "The repo url returned a 404 error. Try using the brute_force flag."
                            )
                            exit()

                    found_source = False
                    source_name = None

                    if "source.zip" in html:
                        found_source = True
                        source_name = "source"

                    if "sources.zip" in html:
                        found_source = True
                        source_name = "sources"

                    # matches = re.findall('\.zip">(.*?)\.zip</a>', html, re.I) # match on text
                    matches = re.findall('href="(.*?)">', html, re.I)  # match on links
                    matches = [
                        m[:-4] for m in matches if m[-4:] == ".zip"
                    ]  # match on zip files and remove extension

                    for m in matches:
                        for os in self.os_names:
                            idx = m.find(os)

                            if idx != -1:
                                arch = m[idx + len(os) :]

                                for after_arch in self.after_archs:
                                    arch = arch.replace(after_arch, "")

                                plats[os].add(arch)
                                break

                # Formatting
                print()
                print(f"Wrapper / static_lib :: {w_name}")
                print(f"Artifact ID :: {art}")
                print(f"URL :: {dir_url}")
                print()

                if found_source:
                    print("This repo appears to provide sources.")
                    print("The name of the source file is:", source_name)
                    print()

                plat_str = "supported_platforms = [\n"
                for os in plats:
                    for arch in plats[os]:
                        plat_str += '    {{ os = "{}", arch = "{}" }},\n'.format(
                            os, arch
                        )
                plat_str += "]"

                print(plat_str)


def main():

    parser = argparse.ArgumentParser(prog="robotpy-build")
    parent_parser = argparse.ArgumentParser(add_help=False)
    subparsers = parser.add_subparsers(dest="cmd")
    subparsers.required = True

    for cls in (
        BuildDep,
        GenCreator,
        HeaderScanner,
        ImportCreator,
        PlatformInfo,
        ShowOverrides,
        MavenParser,
    ):
        cls.add_subparser(parent_parser, subparsers).set_defaults(cls=cls)

    args = parser.parse_args()
    cmd = args.cls()
    retval = cmd.run(args)

    if retval is False:
        retval = 1
    elif retval is True:
        retval = 0
    elif isinstance(retval, int):
        pass
    else:
        retval = 0

    exit(retval)


if __name__ == "__main__":
    main()



from rpytest import ft


def test_defaults():
    assert ft.fnSimpleDefaultParam(1) == 4
    assert ft.fnSimpleDefaultParam(2, 3) == 5

    assert ft.fnEmptyDefaultParam() == []
    assert ft.fnEmptyDefaultParam([1, 2]) == [1, 2]



import inspect
from rpytest import ft


def test_docstrings_enum():

    assert inspect.getdoc(ft.DocEnum) == inspect.cleandoc(
        """
        An enum that is documented
        Maybe it's not great docs, but it's something.

        Members:

          Value1 : value 1 doc
        
          Value2 : value 2 doc??
        """
    )


def test_docstrings_cls():
    assert inspect.getdoc(ft.DocClass) == inspect.cleandoc(
        """
        A class with documentation
        The docs are way cool.
        """
    )


def test_docstrings_meth():
    assert inspect.getdoc(ft.DocClass.fn) == inspect.cleandoc(
        """
        fn(self: rpytest.ft._rpytest_ft.DocClass) -> None
        
        Function with docstring for good measure
        """
    )


def test_docstrings_meth_kwd():
    assert inspect.getdoc(ft.DocClass.fn2) == inspect.cleandoc(
        """
        fn2(self: rpytest.ft._rpytest_ft.DocClass, from_: int) -> None
        
        Function with parameter that's a python keyword

        :param from_: The from parameter
        """
    )


def test_docstrings_meth_rename():
    assert inspect.getdoc(ft.DocClass.fn3) == inspect.cleandoc(
        """
        fn3(self: rpytest.ft._rpytest_ft.DocClass, ohai: int) -> None
        
        Function with renamed parameter

        :param ohai: The renamed parameter
        """
    )


def test_docstrings_var():
    assert (
        inspect.getdoc(ft.DocClass.sweet_var)
        == "An awesome variable, use it for something"
    )


def test_docstrings_fn():
    assert inspect.getdoc(ft.important_retval) == inspect.cleandoc(
        """
        important_retval() -> int

        This function returns something very important
        """
    )


def test_docstrings_append():

    assert inspect.getdoc(ft.DocAppendEnum) == inspect.cleandoc(
        """
        An enum that is documented
        Maybe it's not great docs, but it's something.
        Useful extra information about this enum


        Members:

          Value1 : value 1 doc
          Useful extra information about this value
          
        
          Value2 : value 2 doc??
        """
    )

    assert inspect.getdoc(ft.DocAppendClass) == inspect.cleandoc(
        """
        A class with documentation
        The docs are way cool.
        Useful extra information about this sweet class
        """
    )
    assert inspect.getdoc(ft.DocAppendClass.fn) == inspect.cleandoc(
        """
        fn(self: rpytest.ft._rpytest_ft.DocAppendClass) -> None
        
        Function with docstring for good measure
        Useful extra information about this fn
        """
    )
    assert inspect.getdoc(ft.DocAppendClass.sweet_var) == inspect.cleandoc(
        """
        An awesome variable, use it for something
        Useful extra information about this sweet var
        """
    )


def test_docstrings_template():
    assert inspect.getdoc(ft.DocTemplateSet) == "Set docs to this"

    assert inspect.getdoc(ft.DocTemplateAppend) == inspect.cleandoc(
        """
        A templated class
        Extra appended docs
    """
    )



import rpytest.ft._rpytest_ft as ft


def test_refquals_ref():
    r = ft.RefQuals()
    assert ft.refquals_fn1(r) == 1


def test_refquals_move1():
    r = ft.RefQuals()
    assert ft.refquals_fn2(r) == 2


def test_refquals_move_custom():
    class MyRefQuals(ft.RefQuals):
        pass

    r = MyRefQuals()
    assert ft.refquals_fn2(r) == 42


def test_refquals_move2():
    r = ft.RefQuals()

    # This calls our cpp_code override
    assert r.fn3() == 43

    # This still calls the C++ version since we can't override it
    # merely with cpp_code
    assert ft.refquals_fn3(r) == 3



import importlib.util
import json
import os
from os.path import abspath, exists, dirname, join
import subprocess
import sys

import pybind11_stubgen
from setuptools import Command
from distutils.errors import DistutilsError

from .util import get_install_root


class GeneratePyiError(DistutilsError):
    pass


class BuildPyi(Command):
    base_package: str

    command_name = "build_pyi"
    description = "Generates pyi files from built extensions"

    user_options = [("build-lib=", "d", 'directory to "build" (copy) to')]

    def initialize_options(self):
        self.build_lib = None

    def finalize_options(self):
        self.set_undefined_options("build", ("build_lib", "build_lib"))

    def run(self):
        # cannot build pyi files when cross-compiling
        if (
            "_PYTHON_HOST_PLATFORM" in os.environ
            or "PYTHON_CROSSENV" in os.environ
            or os.environ.get("RPYBUILD_SKIP_PYI") == "1"
        ):
            return

        # Gather information for needed stubs
        data = {"mapping": {}, "stubs": []}

        # OSX-specific: need to set DYLD_LIBRARY_PATH otherwise modules don't
        # work. Luckily, that information was computed when building the
        # extensions...
        env = os.environ.copy()
        dyld_path = set()

        # Requires information from build_ext to work
        build_ext = self.distribution.get_command_obj("build_ext")
        if build_ext.inplace:
            data["out"] = get_install_root(self)
        else:
            data["out"] = self.build_lib

        # Ensure that the associated packages can always be found locally
        for wrapper in build_ext.wrappers:
            pkgdir = wrapper.package_name.split(".")
            init_py = abspath(join(self.build_lib, *pkgdir, "__init__.py"))
            if exists(init_py):
                data["mapping"][wrapper.package_name] = init_py

        # Ensure that the built extension can always be found
        for ext in build_ext.extensions:
            fname = build_ext.get_ext_filename(ext.name)
            data["mapping"][ext.name] = abspath(join(self.build_lib, fname))
            data["stubs"].append(ext.name)

            rpybuild_libs = getattr(ext, "rpybuild_libs", None)
            if rpybuild_libs:
                for pth, _ in rpybuild_libs.values():
                    dyld_path.add(dirname(pth))

        # Don't do anything if nothing is needed
        if not data["stubs"]:
            return

        # OSX-specific
        if dyld_path:
            dyld_path = ":".join(dyld_path)
            if "DYLD_LIBRARY_PATH" in env:
                dyld_path += ":" + env["DYLD_LIBRARY_PATH"]
            env["DYLD_LIBRARY_PATH"] = dyld_path

        data_json = json.dumps(data)

        # Execute in a subprocess in case it crashes
        args = [sys.executable, "-m", __name__]
        try:
            subprocess.run(args, input=data_json.encode("utf-8"), env=env, check=True)
        except subprocess.CalledProcessError:
            raise GeneratePyiError(
                "Failed to generate .pyi file (see above, or set RPYBUILD_SKIP_PYI=1 to ignore) via %s"
                % (args,)
            ) from None

        # Create a py.typed for PEP 561
        with open(join(data["out"], *self.base_package.split("."), "py.typed"), "w"):
            pass


class _PackageFinder:
    """
    Custom loader to allow loading built modules from their location
    in the build directory (as opposed to their install location)
    """

    mapping = {}

    @classmethod
    def find_spec(cls, fullname, path, target=None):
        m = cls.mapping.get(fullname)
        if m:
            return importlib.util.spec_from_file_location(fullname, m)


def generate_pyi(module_name: str, pyi_filename: str):

    print("generating", pyi_filename)

    pybind11_stubgen.FunctionSignature.n_invalid_signatures = 0
    module = pybind11_stubgen.ModuleStubsGenerator(module_name)
    module.parse()
    if pybind11_stubgen.FunctionSignature.n_invalid_signatures > 0:
        print("FAILED to generate pyi for", module_name, file=sys.stderr)
        return False

    module.write_setup_py = False
    with open(pyi_filename, "w") as fp:
        fp.write("#\n# AUTOMATICALLY GENERATED FILE, DO NOT EDIT!\n#\n\n")
        fp.write("\n".join(module.to_lines()))

    typed = join(dirname(pyi_filename), "py.typed")
    print("generating", typed)
    if not exists(typed):
        with open(typed, "w") as fp:
            pass

    return True


def main():

    cfg = json.load(sys.stdin)

    # Configure custom loader
    _PackageFinder.mapping = cfg["mapping"]
    sys.meta_path.insert(0, _PackageFinder)

    # Generate pyi modules
    sys.argv = [
        "<dummy>",
        "--no-setup-py",
        "--log-level=WARNING",
        "--root-module-suffix=",
        "--ignore-invalid",
        "defaultarg",
        "-o",
        cfg["out"],
    ] + cfg["stubs"]

    pybind11_stubgen.main()


if __name__ == "__main__":
    main()



from . import _init_rpytest_dl

# autogenerated by 'robotpy-build create-imports rpytest.dl rpytest.dl._rpytest_dl'
from ._rpytest_dl import downloaded_fn, extra_content

__all__ = ["downloaded_fn", "extra_content"]



import atexit
import contextlib
import os
from os.path import dirname, exists, join, normpath
import posixpath
import shutil
import sys
import urllib.request
import tempfile
import zipfile

from .version import version


USER_AGENT = f"robotpy-build/{version}"
SHOW_PROGRESS = "CI" not in os.environ


def _download(url: str, dst_fname: str):
    """
    Downloads a file to a specified directory
    """

    def _reporthook(count, blocksize, totalsize):
        if SHOW_PROGRESS:
            percent = int(count * blocksize * 100 / totalsize)
            sys.stdout.write("\r%02d%%" % percent)
            sys.stdout.flush()

    print("Downloading", url)

    request = urllib.request.Request(url, headers={"User-Agent": USER_AGENT})

    with contextlib.closing(urllib.request.urlopen(request)) as fp:
        headers = fp.info()

        with open(dst_fname, "wb") as tfp:

            # copied from urlretrieve source code, Python license
            bs = 1024 * 8
            size = -1
            blocknum = 0
            read = 0
            if "content-length" in headers:
                size = int(headers["Content-Length"])

            while True:
                block = fp.read(bs)
                if not block:
                    break
                read += len(block)
                tfp.write(block)
                blocknum += 1
                if _reporthook:
                    _reporthook(blocknum, bs, size)

    if SHOW_PROGRESS:
        sys.stdout.write("\n")
        sys.stdout.flush()


def download_and_extract_zip(url, to, cache):
    """
    Utility method intended to be useful for downloading/extracting
    third party source zipfiles

    :param to: is either a string or a dict of {src: dst}
    """

    os.makedirs(cache, exist_ok=True)
    zip_fname = join(cache, posixpath.basename(url))
    if not exists(zip_fname):
        _download(url, zip_fname)

    with zipfile.ZipFile(zip_fname) as z:
        if isinstance(to, str):
            to = {"": to}

        for src, dst in to.items():
            if src == "":
                z.extractall(dst)
            else:
                # if is directory, copy whole thing recursively
                try:
                    info = z.getinfo(src)
                except KeyError as e:
                    osrc = src
                    src = src + "/"
                    try:
                        info = z.getinfo(src)
                    except KeyError:
                        info = None
                    if info is None:
                        msg = f"error extracting {osrc} from {zip_fname}"
                        raise ValueError(msg) from e
                if info.is_dir():
                    ilen = len(info.filename)
                    for minfo in z.infolist():
                        if minfo.is_dir():
                            continue
                        srcname = posixpath.normpath(minfo.filename)
                        if srcname.startswith(info.filename):
                            dstname = join(dst, normpath(srcname[ilen:]))
                            dstdir = dirname(dstname)
                            if not exists(dstdir):
                                os.makedirs(dstdir)
                            with z.open(minfo.filename, "r") as zfp, open(
                                dstname, "wb"
                            ) as fp:
                                shutil.copyfileobj(zfp, fp)
                else:
                    # otherwise write a single file
                    with z.open(src, "r") as zfp, open(dst, "wb") as fp:
                        shutil.copyfileobj(zfp, fp)



from rpytest.ft import LTTester, LTWithVirtual
import gc


class PyLTWithVirtual(LTWithVirtual):
    def get_bool(self) -> bool:
        return True


def test_lifetime():
    tester = LTTester()

    # At this point python releases the reference to the created object
    # need to ensure that it stays around as long as the std::shared_ptr
    tester.set_val(PyLTWithVirtual())
    gc.collect()
    assert tester.get_bool() == True



from . import _init_rpytest_ft

# autogenerated by 'robotpy-build create-imports rpytest.ft rpytest.ft._rpytest_ft'
from ._rpytest_ft import (
    Abstract,
    ClassWithFields,
    ClassWithIgnored,
    ClassWithTrampoline,
    DocAppendClass,
    DocAppendEnum,
    DocClass,
    DocEnum,
    DocTemplateAppend,
    DocTemplateSet,
    EnumContainer,
    EnumContainer2,
    EnumWithIgnored,
    GCEnum,
    GEnum,
    GEnumMath,
    HasFactory,
    IBase,
    IChild,
    IFinal,
    IGrandChild,
    IMChild,
    IMOther,
    InlineCode,
    LTTester,
    LTWithVirtual,
    MVB,
    MVC,
    MVD0,
    MVD1,
    MVE,
    MVF,
    NSEnumContainer,
    NSGCEnum,
    NSGEnum,
    NestedTypecaster,
    Nurse,
    OB,
    OBInitializer,
    OBinitOB,
    OC,
    OCInitializer,
    OCinitOB,
    OCinitOC,
    OG,
    OGinitOC,
    OverloadedObject,
    PBase,
    PChild,
    PGChild,
    PatientRef,
    RenamedClass,
    RenamedEnum,
    StaticOnly,
    StripPrefixEnum,
    TBase,
    TBaseGetN4,
    TBaseGetN6,
    TBasicString,
    TChildGetN4,
    TChildGetN6,
    TClassWithFn,
    TConcrete,
    TDependentParamInt,
    TDependentUsingInt,
    TDependentUsing2Int,
    TOuter,
    TcrtpConcrete,
    TcrtpFwdConcrete,
    UPBase,
    UPChild,
    VBase,
    VChild,
    VirtualComma,
    checkConvertRpyintToInt,
    check_impure_io,
    check_pure_io,
    convertRpyintToInt,
    fnEmptyDefaultParam,
    fnIgnoredParam,
    fnOverload,
    fnParamArrayOut,
    fnParamArrayOutWithDefault,
    fnParamFundConstRef,
    fnParamFundPtr,
    fnParamFundRef,
    fnRenamed,
    fnRenamedParam,
    fnSimpleDefaultParam,
    get123,
    important_retval,
    raise_from,
    raise_from_already_set,
    subpkg,
)

__all__ = [
    "Abstract",
    "ClassWithFields",
    "ClassWithIgnored",
    "ClassWithTrampoline",
    "DocAppendClass",
    "DocAppendEnum",
    "DocClass",
    "DocEnum",
    "DocTemplateAppend",
    "DocTemplateSet",
    "EnumContainer",
    "EnumContainer2",
    "EnumWithIgnored",
    "GCEnum",
    "GEnum",
    "GEnumMath",
    "HasFactory",
    "IBase",
    "IChild",
    "IFinal",
    "IGrandChild",
    "IMChild",
    "IMOther",
    "InlineCode",
    "LTTester",
    "LTWithVirtual",
    "MVB",
    "MVC",
    "MVD0",
    "MVD1",
    "MVE",
    "MVF",
    "NSEnumContainer",
    "NSGCEnum",
    "NSGEnum",
    "NestedTypecaster",
    "Nurse",
    "OB",
    "OBInitializer",
    "OBinitOB",
    "OC",
    "OCInitializer",
    "OCinitOB",
    "OCinitOC",
    "OG",
    "OGinitOC",
    "OverloadedObject",
    "PBase",
    "PChild",
    "PGChild",
    "PatientRef",
    "RenamedClass",
    "RenamedEnum",
    "StaticOnly",
    "StripPrefixEnum",
    "TBase",
    "TBaseGetN4",
    "TBaseGetN6",
    "TBasicString",
    "TChildGetN4",
    "TChildGetN6",
    "TClassWithFn",
    "TConcrete",
    "TDependentParamInt",
    "TDependentUsingInt",
    "TDependentUsing2Int",
    "TOuter",
    "TcrtpConcrete",
    "TcrtpFwdConcrete",
    "UPBase",
    "UPChild",
    "VBase",
    "VChild",
    "VirtualComma",
    "checkConvertRpyintToInt",
    "check_impure_io",
    "check_pure_io",
    "convertRpyintToInt",
    "fnEmptyDefaultParam",
    "fnIgnoredParam",
    "fnOverload",
    "fnParamArrayOut",
    "fnParamArrayOutWithDefault",
    "fnParamFundConstRef",
    "fnParamFundPtr",
    "fnParamFundRef",
    "fnRenamed",
    "fnRenamedParam",
    "fnSimpleDefaultParam",
    "get123",
    "important_retval",
    "raise_from",
    "raise_from_already_set",
    "subpkg",
]



# yes, we include some stdint types in here, but that's fine, this
# is just a best effort generator
_builtins = {
    "void": "v",
    "wchar_t": "w",
    "bool": "b",
    "char": "c",
    "signed char": "a",
    "int8_t": "a",
    "unsigned char": "h",
    "uint8_t": "c",
    "short": "s",
    "int16_t": "s",
    "unsigned short": "t",
    "uint16_t": "t",
    "int": "i",
    "int32_t": "i",
    "unsigned int": "j",
    "uint32_t": "j",
    "long": "l",
    "unsigned long": "m",
    "long long": "x",
    "__int64": "x",
    "int64_t": "x",
    "unsigned long long": "y",
    "unsigned __int64": "y",
    "uint64_t": "y",
    "__int128": "n",
    "unsigned __int128": "o",
    "float": "f",
    "double": "d",
    "long double": "e",
    "__float128": "g",
    "char32_t": "Di",
    "char16_t": "Ds",
    "auto": "Da",
}

_type_bad_chars = ":<>=()&,"
_type_trans = str.maketrans(_type_bad_chars, "_" * len(_type_bad_chars))


def _encode_type(param):
    names = []

    # prefix with cv-qualifiers, refs, pointers
    if param.get("constant"):
        names.append("K")
    if param.get("volatile"):
        names.append("V")

    if param.get("array"):
        names.append("A" * param.get("array"))

    refs = param["reference"]
    if refs == 1:
        names.append("R")
    elif refs == 2:
        names.append("O")

    ptr = param["pointer"]
    if ptr:
        names.append("P" * ptr)

    # actual type
    raw_type = param.get("enum", param["raw_type"])
    typ = _builtins.get(raw_type)
    if not typ:
        # Only mangle the typename, ignore namespaces as children might have the types
        # aliased or something. There are cases where this would fail, but hopefully
        # that doesn't happen?
        # TODO: do this better
        raw_type = raw_type.split("::")[-1]
        # assert " " not in raw_type, raw_type
        typ = "T" + raw_type.replace(" ", "").translate(_type_trans)

    names.append(typ)
    return "".join(names)


def trampoline_signature(fn):
    """
    In our trampoline functions, each function can be disabled by defining
    a macro corresponding to the function type. This helper function
    generates a mangled name based on the function name and types.

    This is roughly based on the IA64 name mangling scheme that GCC uses,
    but omits template stuff.

    function name
    cv qualifiers
    parameter types
    """

    # TODO: operator overloads
    names = []

    if fn["const"]:
        names.append("K")
    refqual = fn["ref_qualifiers"]
    if refqual:
        if refqual == "&":
            names.append("R")
        if refqual == "&&":
            names.append("O")

    names.append(fn["name"])

    params = fn["parameters"]
    if not params:
        names.append("_v")
    else:
        for p in params:
            names.append("_")
            names.append(_encode_type(p))

    if fn["vararg"]:
        names.append("_z")

    return "".join(names)



from rpytest import ft


def test_fn_overloads():
    assert ft.fnOverload(1) == 1
    assert ft.fnOverload(1, 2) == 2


def test_cls_overloads():
    oo = ft.OverloadedObject()
    assert oo.overloaded(1) == 0x1
    assert oo.overloaded("bob") == 0x2

    assert oo.overloaded_constexpr(1, 2) == 3
    assert oo.overloaded_constexpr(1, 2, 3) == 6

    assert ft.OverloadedObject.overloaded_static(1) == 0x3
    assert ft.OverloadedObject.overloaded_static("yup") == 0x4



import typing

from .pyproject_configs import Download, MavenLibDownload


def _get_artifact_url(dlcfg: MavenLibDownload, classifier: str) -> str:
    # TODO: support development against locally installed things?
    repo_url = dlcfg.repo_url
    grp = dlcfg.group_id.replace(".", "/")
    art = dlcfg.artifact_id
    ver = dlcfg.version

    return f"{repo_url}/{grp}/{art}/{ver}/{art}-{ver}-{classifier}.zip"


def convert_maven_to_downloads(
    mcfg: MavenLibDownload, static: bool
) -> typing.List[Download]:
    """
    Converts a MavenLibDownload object to a list of normal downloads
    """

    dl_lib = {}
    dl_header = {}
    dl_sources = {}

    if mcfg.use_sources:
        if static:
            raise ValueError("Cannot specify sources in static_lib section")

        # sources don't have libs, ignore them
        dl_sources["url"] = _get_artifact_url(mcfg, mcfg.sources_classifier)
        dl_sources["sources"] = mcfg.sources
        dl_sources["patches"] = mcfg.patches
    elif mcfg.sources is not None:
        raise ValueError("sources must be None if use_sources is False!")
    elif mcfg.patches is not None:
        raise ValueError("patches must be None if use_sources is False!")
    else:
        # libs

        dl_lib["libs"] = mcfg.libs
        if mcfg.libs is mcfg.dlopenlibs is None:
            dl_lib["libs"] = [mcfg.artifact_id]
        dl_lib["dlopenlibs"] = mcfg.dlopenlibs
        dl_lib["libexts"] = mcfg.libexts
        dl_lib["linkexts"] = mcfg.linkexts

        if static:
            dl_lib["libdir"] = "{{ OS }}/{{ ARCH }}/static"
            dl_lib["url"] = _get_artifact_url(mcfg, "{{ OS }}{{ ARCH }}static")
        else:
            dl_lib["libdir"] = "{{ OS }}/{{ ARCH }}/shared"
            dl_lib["url"] = _get_artifact_url(mcfg, "{{ OS }}{{ ARCH }}")

    # headers
    dl_header["incdir"] = ""
    dl_header["url"] = _get_artifact_url(mcfg, "headers")
    dl_header["header_patches"] = mcfg.header_patches

    # Construct downloads and return it
    downloads = []
    for d in (dl_lib, dl_header, dl_sources):
        if d:
            downloads.append(Download(**d))

    return downloads



"""
    On OSX, the loader does not look at the current process to load
    dylibs -- it insists on finding them itself, so we have to fixup
    our binaries such that they resolve correctly.
    
    Two cases we need to deal with
    - Local development/installation
    - Building a wheel for pypi
    
    In development, we assume things are installed exactly where they will be
    at runtime.
      -> @loader_path/{relpath(final_location, dep_path)}
    
    For pypi wheels, we assume that installation is in site-packages, and
    so are the libraries that this lib depends on.
      -> @loader_path/{relpath(final_siterel, dep_siterel)}
    
    Notice these are the same IF you only build wheels in a virtualenv
    that only has its dependencies installed in site-packages
    

    .. warning:: This will only work for the environment it's compiled in!
                 This basically means don't compile wheels in your development 
                 environment, use a clean environment instead

"""


from delocate.delocating import filter_system_libs
from delocate.tools import get_install_names, set_install_name as _set_install_name

from os import path

from typing import Dict, List, Optional, Tuple

from .pkgcfg_provider import PkgCfg, PkgCfgProvider


def set_install_name(file: str, old_install_name: str, new_install_name: str):
    """Change the install name for a library

    :param file: path to a executable/library file
    :param old_install_name: current path to dependency
    :param new_install_name: new path to dependency
    """

    # This function just calls delocate's set_install_name which uses install_name_tool.
    # This function exists in case we want to change the implementation.

    _set_install_name(file, old_install_name, new_install_name)
    print("Relink:", file, ":", old_install_name, "->", new_install_name)


# Common data structure used here
# - key is basename of library file
# - Value tuple has two pieces:
#   - 0: Where the library file really is right now
#   - 1: Where the library file will be when installed
LibsDict = Dict[str, Tuple[str, str]]


def _resolve_libs(libpaths: Optional[List[str]], libname_full: str, libs: LibsDict):
    if not libpaths:
        return
    for libpath in libpaths:
        p = path.join(libpath, libname_full)
        if path.exists(p):
            libs[libname_full] = (p, p)
            return


def _resolve_libs_in_self(dep: PkgCfg, install_root: str, libs: LibsDict):
    pkgroot = path.join(install_root, *dep.package_name.split("."))
    full_names = dep.get_library_full_names()
    if not full_names:
        return
    for libname_full in full_names:
        for ld, ldr in zip(dep.get_library_dirs(), dep.get_library_dirs_rel()):
            p = path.join(ld, libname_full)
            if path.exists(p):
                # stores where it will exist
                libs[libname_full] = (p, path.join(pkgroot, ldr, libname_full))
                break


def _resolve_dependencies(
    install_root: str, pkg: PkgCfg, pkgcfg: PkgCfgProvider, libs: LibsDict
):
    # first, gather all possible libraries by retrieving this package and
    # it's dependents. We're not concerned about redirecting non-robotpy-build
    # libraries, since we can't control where those are located
    deps = pkgcfg.get_all_deps(pkg.name)

    pypi_package = pkg.pypi_package

    for dep in deps:
        # dependencies are in their installed location
        # .. except when they're in the same wheel
        if pypi_package and dep.pypi_package == pypi_package:
            _resolve_libs_in_self(dep, install_root, libs)
        else:
            libdirs = dep.get_library_dirs()
            full_names = dep.get_library_full_names()
            if full_names:
                for libname_full in full_names:
                    _resolve_libs(libdirs, libname_full, libs)


def _fix_libs(to_fix: LibsDict, libs: LibsDict):

    for (current_libpath, install_libpath) in to_fix.values():
        for lib in get_install_names(current_libpath):
            libb = path.basename(lib)
            libdata = libs.get(libb)
            if libdata:
                desired_path = path.relpath(libdata[1], path.dirname(install_libpath))
                desired_path = "@loader_path/" + desired_path
                set_install_name(current_libpath, lib, desired_path)
            elif filter_system_libs(lib):
                raise ValueError(
                    "unresolved lib %s: maybe a dependency is missing?" % lib
                )


def relink_libs(install_root: str, pkg: PkgCfg, pkgcfg: PkgCfgProvider):
    """
    Given a package, relink it's external libraries

    :param install_root: Where this package will be (is) installed
    :param pkg: Object that implements pkgcfg for this wrapper
    :param pkgcfg: robotpy-build pkgcfg resolver
    """
    libs: LibsDict = {}
    _resolve_dependencies(install_root, pkg, pkgcfg, libs)
    to_fix: LibsDict = {}
    _resolve_libs_in_self(pkg, install_root, to_fix)
    libs.update(to_fix)
    _fix_libs(to_fix, libs)


def relink_extension(
    install_root: str,
    extension_path: str,
    extension_rel: str,
    pkg: PkgCfg,
    pkgcfg: PkgCfgProvider,
) -> LibsDict:
    """
    Given an extension, relink it

    :param install_root: Where this package will be (is) installed
    :param extension_path: full path to extension library
    :param extension_rel: Relative path to library where it will be (is) installed
    :param pkg: Object that implements pkgcfg for this wrapper
    :param pkgcfg: robotpy-build pkgcfg resolver
    """
    libs: LibsDict = {}
    _resolve_dependencies(install_root, pkg, pkgcfg, libs)
    _resolve_libs_in_self(pkg, install_root, libs)

    to_fix = {
        path.basename(extension_path): (
            extension_path,
            path.join(install_root, extension_rel),
        )
    }
    _fix_libs(to_fix, libs)
    return libs



from rpytest.ft import ClassWithTrampoline


def test_class_with_trampoline():
    c = ClassWithTrampoline()
    assert c.get42() == 42

    assert c.fnWithMoveOnlyParam(4) == 4
    assert ClassWithTrampoline.check_moveonly(c) == 7


def test_trampoline_with_mv():
    class PyClassWithTrampoline(ClassWithTrampoline):
        pass

    c = PyClassWithTrampoline()
    assert ClassWithTrampoline.check_moveonly(c) == 7


def test_trampoline_without_mv():
    class PyClassWithTrampoline(ClassWithTrampoline):
        def fnWithMoveOnlyParam(self, i):
            assert i == 7
            return 8

    c = PyClassWithTrampoline()
    assert ClassWithTrampoline.check_moveonly(c) == 8



import os
from typing import Optional, List
import yaml

from .hooks_datacfg import Model


class DevConfig(Model):
    """
    Configuration options useful for developing robotpy-build wrappers.
    To use these, set the environment variable RPYBUILD_GEN_FILTER=filename.yml
    """

    #: When set this will only generate new wrapping code for the specified
    #: headers (left side of generate in pyproject.toml). Existing wrapping
    #: code will not be deleted during a build.
    #:
    #: Useful in conjunction with ccache
    only_generate: Optional[List[str]] = None


def get_dev_config(name: str) -> Optional[DevConfig]:
    # name is the wrapper config, not used currently
    genfilter = os.environ.get("RPYBUILD_GEN_FILTER")
    if genfilter:
        with open(genfilter) as fp:
            data = yaml.safe_load(fp)
    else:
        data = {}

    return DevConfig(**data)



from rpytest import ft


def test_array_out():
    assert ft.fnParamArrayOut() == (4, [5, 6, 7])


def test_array_out_w_default():
    assert ft.fnParamArrayOutWithDefault() == (4, [0, 6, 2])


def test_fund_ptr():
    assert ft.fnParamFundPtr(4) == (3, 5)


def test_fund_ref():
    assert ft.fnParamFundRef(6) == (7, 5)


def test_fund_const_ref():
    assert ft.fnParamFundConstRef(1, 2) == 3



#
# pyproject.toml
#

import os
import re
from typing import Dict, List, Optional

from .util import Model

_arch_re = re.compile(r"\{\{\s*ARCH\s*\}\}")
_os_re = re.compile(r"\{\{\s*OS\s*\}\}")


class PatchInfo(Model):
    """
    A unified diff to apply to downloaded source code before building a
    a wrapper.

    .. code-block:: toml

       [[tool.robotpy-build.wrappers."MY.PACKAGE.NAME".maven_lib_download.patches]]
       patch = "path/to/my.patch"
       strip = 0
    """

    #: Name of patch file to apply
    patch: str

    #: Number of directories to strip
    strip: int = 0


class MavenLibDownload(Model):
    """
    Used to download artifacts from a maven repository. This can download
    headers, shared libraries, and sources.

    .. code-block:: toml

       [tool.robotpy-build.wrappers."PACKAGENAME".maven_lib_download]
       artifact_id = "mything"
       group_id = "com.example.thing"
       repo_url = "http://example.com/maven"
       version = "1.2.3"

    .. note:: For FIRST Robotics libraries, the information required can
              be found in the vendor JSON file
    """

    #: Maven artifact ID
    artifact_id: str

    #: Maven group ID
    group_id: str

    #: Maven repository URL
    repo_url: str

    #: Version of artifact to download
    version: str

    #: Configure the sources classifier
    sources_classifier: str = "sources"

    #: When set, download sources instead of downloading libraries. When
    #: using this, you need to manually add the sources to the configuration
    #: to be compiled via :attr:`sources`.
    use_sources: bool = False

    # common with Download

    #: Names of contained shared libraries (in loading order). If None,
    #: set to artifact_id.
    libs: Optional[List[str]] = None

    #: Names of contained shared link only libraries (in loading order). If None,
    #: set to name. If empty list, link only libs will not be downloaded.
    dlopenlibs: Optional[List[str]] = None

    #: Library extensions map
    libexts: Dict[str, str] = {}

    #: Compile time extensions map
    linkexts: Dict[str, str] = {}

    #: If :attr:`use_sources` is set, this is the list of sources to compile
    sources: Optional[List[str]] = None

    #: If :attr:`use_sources` is set, apply the following patches to the sources. Patches
    #: must be in unified diff format.
    patches: Optional[List[PatchInfo]] = None

    #: Patches to downloaded header files. Patches must be in unified diff format.
    header_patches: Optional[List[PatchInfo]] = None


class Download(Model):
    """
    Download sources/libs/includes from a single file

    .. code-block:: toml

       [[tool.robotpy-build.wrappers."PACKAGENAME".download]]
       url = "https://my/url/something.zip"
       incdir = "include"
       libs = ["mylib"]

    """

    #: URL of zipfile to download
    #:
    #: {{ARCH}} and {{OS}} are replaced with the architecture/os name
    url: str

    #: Directory that contains include files.
    #:
    #: {{ARCH}} and {{OS}} are replaced with the architecture/os name
    incdir: Optional[str] = None

    #: Directory that contains library files
    #:
    #: {{ARCH}} and {{OS}} are replaced with the architecture/os name
    libdir: str = ""

    #: Extra include paths, relative to the include directory
    #:
    #: {{ARCH}} and {{OS}} are replaced with the architecture/os name
    extra_includes: List[str] = []

    # Common with MavenLibDownload

    #: If specified, names of contained shared libraries (in loading order)
    libs: Optional[List[str]] = None

    #: If specified, names of contained shared link only libraries (in loading order).
    #: If None, set to name. If empty list, link only libs will not be downloaded.
    dlopenlibs: Optional[List[str]] = None

    #: Library extensions map
    libexts: Dict[str, str] = {}

    #: Compile time extensions map
    linkexts: Dict[str, str] = {}

    #: List of sources to compile
    sources: Optional[List[str]] = None

    #: If :attr:`sources` is set, apply the following patches to the sources. Patches
    #: must be in unified diff format.
    patches: Optional[List[PatchInfo]] = None

    #: Patches to downloaded header files in incdir. Patches must be in unified
    #: diff format.
    header_patches: Optional[List[PatchInfo]] = None

    def _update_with_platform(self, platform):
        for n in ("url", "incdir", "libdir"):
            v = getattr(self, n, None)
            if v is not None:
                v = _os_re.sub(platform.os, _arch_re.sub(platform.arch, v))
                setattr(self, n, v)

        if self.extra_includes:
            self.extra_includes = [
                _os_re.sub(platform.os, _arch_re.sub(platform.arch, v))
                for v in self.extra_includes
            ]


class StaticLibConfig(Model):
    """
    Static libraries that can be consumed as a dependency by other wrappers
    in the same project. Static libraries are not directly installed, and
    as a result cannot be consumed by other projects.

    .. code-block:: toml

       [tool.robotpy-build.static_libs."MY.PACKAGE.NAME"]

    """

    #: If this project depends on external libraries stored in a maven repo
    #: specify it here
    maven_lib_download: Optional[MavenLibDownload] = None

    #: If this project depends on external libraries downloadable from some URL
    #: specify it here
    download: Optional[List[Download]] = None

    #: If True, skip this library; typically used in conjection with an override
    ignore: bool = False


class TypeCasterConfig(Model):
    """
    Specifies type casters that this package exports. robotpy-build
    will attempt to detect these types at generation time and include
    them in generated wrappers.

    .. code-block:: toml

       [[tool.robotpy-build.wrappers."PACKAGENAME".type_casters]]
       header = "my_type_caster.h"
       types = ["foo_t", "ns::ins::bar_t"]

    .. seealso:: :ref:`type_casters`
    """

    #: Header file to include when one of the types are detected in a wrapper
    header: str

    #: Types to look for to indicate that this type caster header should be
    #: included.
    types: List[str]

    #: If a parameter type that requires this type caster requires a default
    #: argument, a C-style ``(type)`` cast is used on the default argument.
    #:
    #: The default cast can be disabled via param_override's ``disable_type_caster_default_cast``
    default_arg_cast: bool = False


class WrapperConfig(Model):
    """
    Configuration for building a C++ python extension module, optionally
    using autogenerated wrappers around existing library code.

    .. code-block:: toml

       [tool.robotpy-build.wrappers."PACKAGENAME"]
       name = "package_name"

    The PACKAGENAME above is a python package (eg "example.package.name").
    A robotpy-build project can contain many different wrappers and packages.
    """

    #: Name that other projects/wrappers use in their 'depends' list
    name: str

    #: Name of extension to build. If None, set to _{name}
    extension: Optional[str] = None

    #: Name of generated file that ensures the shared libraries and any
    #: dependencies are loaded. Defaults to ``_init{extension}.py``
    #:
    #: Generally, you should create an ``__init__.py`` file that imports
    #: this module, otherwise your users will need to do so.
    libinit: Optional[str] = None

    #: List of robotpy-build library dependencies. This affects this wrapper
    #: library in the following ways:
    #:
    #: * Any include file directories exported by the dependency will be added
    #:   to the include path for any source files compiled by this wrapper
    #: * It will be linked to any libraries the dependency contains
    #: * The python module for the dependency will be imported in the
    #:   ``_init{extension}.py`` file.
    depends: List[str] = []

    #: If this project depends on external libraries stored in a maven repo
    #: specify it here.
    maven_lib_download: Optional[MavenLibDownload] = None

    #: If this project depends on external libraries downloadable from some URL
    #: specify it here
    download: Optional[List[Download]] = None

    #: List of extra include directories to export, relative to the
    #: project root.
    extra_includes: List[str] = []

    #: Optional source files to compile. Path is relative to the root of
    #: the project.
    sources: List[str] = []

    #: Specifies header files that autogenerated pybind11 wrappers will be
    #: created for. Simple C++ headers will most likely 'just work', but
    #: complex headers will need to have an accompanying :attr:`generation_data`
    #: file specified that can customize the autogenerated files.
    #:
    #: List of dictionaries: each dictionary key is used for the function
    #: name of the initialization function, the value is the header that is
    #: being wrapped. The header is first looked for relative to the
    #: package, then relative to each include directory (including
    #: downloaded and extracted packages).
    #:
    #: .. code-block:: toml
    #:
    #:    [tool.robotpy-build.wrappers."PACKAGENAME".autogen_headers]
    #:    Name = "header.h"
    #:
    #: .. seealso:: :ref:`autowrap`
    #:
    autogen_headers: Optional[Dict[str, str]] = None

    #: DEPRECATED: Same as autogen_headers, but more complicated
    generate: Optional[List[Dict[str, str]]] = None

    #: Path to a single data.yml to use during code generation, or a directory
    #: of yaml files. If a directory, generation data will be looked up
    #: using the key in the generate dictionary.
    #:
    #: These YAML files can be generated via the robotpy-build command line tool:
    #:
    #: .. code-block:: sh
    #:
    #:    robotpy-build create-gen --write
    #:
    #: .. seealso:: :ref:`gendata`
    #:
    generation_data: Optional[str] = None

    #: Specifies type casters that this package exports.
    type_casters: List[TypeCasterConfig] = []

    #: Preprocessor definitions to apply when compiling this wrapper.
    pp_defines: List[str] = []

    #: If True, skip this wrapper; typically used in conjection with an override.
    ignore: bool = False


class DistutilsMetadata(Model):
    """
    Configures the metadata that robotpy-build passes to setuptools when
    the project is installed. The keys in this section match the standard
    arguments passed to the ``setuptools.setup`` function.

    .. code-block:: toml

       [tool.robotpy-build.metadata]
       name = "my-awesome-dist"
       description = "Cool thing"
       license = "MIT"

    robotpy-build will automatically detect/set the following keys:

    * cmdclass
    * ext_modules
    * include_package_data - ``True``
    * long_description - Contents of README.md/README.rst
    * long_description_content_type - If required
    * packages
    * python_requires - ``>=3.6``
    * version - via setuptools_scm
    * zip_safe - ``False``

    .. note:: This section is required
    """

    class Config:
        # allow passing in extra keywords to setuptools
        extra = "allow"

    #: The name of the package
    name: str

    #: A single line describing the package
    description: Optional[str] = None

    #: The name of the package author
    author: str

    #: The email address of the package author
    author_email: str

    #: A URL for the package (homepage)
    url: str

    #: The license for the package
    license: str

    #: A string or list of strings specifying what other distributions need
    #: to be installed when this one is.
    install_requires: List[str]


class SupportedPlatform(Model):
    """
    Supported platforms for this project. Currently this information is
    merely advisory, and is used to generate error messages when platform
    specific downloads fail.

    .. code-block:: toml

       [tool.robotpy-build]
       base_package = "my.package"
       supported_platforms = [
           { os = "windows", arch = "x86-64" },
       ]

    .. seealso:: List of supported :ref:`platforms <platforms>`

    """

    #: Platform operating system name
    os: Optional[str] = None

    #: Platform architecture
    arch: Optional[str] = None


class RobotpyBuildConfig(Model):
    """
    Contains information for configuring the project

    .. code-block:: toml

       [tool.robotpy-build]
       base_package = "my.package"

    .. note:: This section is required
    """

    #: Python package to store version information and robotpy-build metadata in
    base_package: str

    #:
    #: .. seealso:: :class:`.SupportedPlatform`
    #:
    supported_platforms: List[SupportedPlatform] = []

    #
    # These are all documented in their class, it's more confusing to document
    # them here too.
    #

    metadata: DistutilsMetadata

    wrappers: Dict[str, WrapperConfig] = {}

    static_libs: Dict[str, StaticLibConfig] = {}



from rpytest import ft
import pytest

#
# main module
#


def test_raise_from():
    with pytest.raises(ValueError) as excinfo:
        ft.raise_from()
    assert str(excinfo.value) == "outer"
    assert str(excinfo.value.__cause__) == "inner"


def test_raise_from_already_set():
    with pytest.raises(ValueError) as excinfo:
        ft.raise_from_already_set()
    assert str(excinfo.value) == "outer"
    assert str(excinfo.value.__cause__) == "inner"


#
# abstract.h
#


class MyBadAbstract(ft.Abstract):
    pass


def test_bad_abstract():
    m = MyBadAbstract()

    with pytest.raises(RuntimeError):
        m.mustOverrideMe()


class MyGoodAbstract(ft.Abstract):
    def mustOverrideMe(self):
        return 0x3


def test_good_abstract():
    m = MyGoodAbstract()
    assert m.mustOverrideMe() == 0x3


#
# factory.h
#


def test_factory():
    o = ft.HasFactory(4)
    assert o.m_x == 5


#
# inline_code.h
#


def test_inline_code():
    o = ft.InlineCode()
    assert o.get2() == 2
    assert o.get4() == 4


#
# static_only.h
#


def test_static_only():

    # shouldn't be able to construct
    with pytest.raises(TypeError):
        ft.StaticOnly()

    # should be able to call static
    assert ft.StaticOnly.callme() == 0x56


#
# virtual_xform.h
#
# Check to see that python and C++ see different things for each type
# - py: () -> str
# - C++ (stringstream&) -> void
#


def test_virtual_xform():

    base = ft.VBase()

    with pytest.raises(RuntimeError):
        assert base.pure_io() == "ohai"

    with pytest.raises(RuntimeError):
        assert ft.check_pure_io(base)

    assert base.impure_io() == "py vbase impure + c++ vbase impure"
    assert ft.check_impure_io(base) == "c++ vbase impure"

    class PyChild(ft.VBase):
        def pure_io(self) -> str:
            return "pychild pure"

        def impure_io(self) -> str:
            return "pychild impure"

    pychild = PyChild()
    assert pychild.pure_io() == "pychild pure"
    assert ft.check_pure_io(pychild) == "vbase-xform-pure pychild pure"

    assert pychild.impure_io() == "pychild impure"
    assert ft.check_impure_io(pychild) == "vbase-xform-impure pychild impure"

    child = ft.VChild()

    assert child.pure_io() == "py vchild pure + c++ vchild pure"
    assert ft.check_pure_io(child) == "c++ vchild pure"

    assert child.impure_io() == "py vchild impure + c++ vchild impure"
    assert ft.check_impure_io(child) == "c++ vchild impure"

    class PyGrandChild(ft.VChild):
        def pure_io(self) -> str:
            return "pygrandchild pure"

        def impure_io(self) -> str:
            return "pygrandchild impure"

    pygrandchild = PyGrandChild()

    assert pygrandchild.pure_io() == "pygrandchild pure"
    assert ft.check_pure_io(pygrandchild) == "vchild-xform-pure pygrandchild pure"

    assert pygrandchild.impure_io() == "pygrandchild impure"
    assert ft.check_impure_io(pygrandchild) == "vchild-xform-impure pygrandchild impure"


#
# Misc
#

# ensure that not calling __init__ from a inherited class raises TypeError
def test_init_raises():
    called = [False]

    class PyGoodAbstract(ft.Abstract):
        def __init__(self):
            called[0] = True

    with pytest.raises(TypeError):
        PyGoodAbstract()
    assert called == [True]


def test_subpkg():
    from rpytest.ft.subpkg import SPClass



#!/usr/bin/env python3

import os
from os.path import abspath, dirname, join
import sys
import subprocess

if __name__ == "__main__":

    root = abspath(dirname(__file__))
    os.chdir(root)

    # Install cpp project
    wcpp = join(root, "cpp", "run_install.py")
    subprocess.check_call([sys.executable, wcpp] + sys.argv[1:])

    # Run pytest
    subprocess.check_call([sys.executable, "-m", "pytest"])
